%  Copyright (C) 2002-2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software Foundation,
%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
\section{darcs get}
\begin{code}
module Get ( get ) where

import IO ( hFlush, stdout )
import Directory ( setCurrentDirectory, doesDirectoryExist, doesFileExist,
                   createDirectory,
                 )
import Workaround ( getCurrentDirectory )
import System ( ExitCode(..), exitWith )
import Monad ( when, unless )
import Maybe ( isJust )

import DarcsCommands ( DarcsCommand(..), nodefaults )
import DarcsArguments ( DarcsFlag( TagName, PatchName, RepoName, Partial,
                                   Verbose, Quiet ),
                        any_verbosity, partial, reponame, tagname,
                        patchname_option,
                      )
import Repository ( PatchSet, read_repo, write_inventory,
                    write_checkpoint_patch, mmap_slurp_all_but_darcs,
                    slurp_recorded, is_repo, absolute_dir, get_checkpoint,
                    copy_repo_patches, apply_patches,
                  )
import Patch ( Patch, commute, writePatch, apply_to_slurpy, patch2patchinfo )
import PatchInfo ( PatchInfo, just_name, make_filename )
import SlurpDirectory ( slurp_write, slurp_write_dirty, empty_slurpy )
import Workaround ( mkRegex, matchRegex )
import External ( copyFileOrUrl )
import Depends ( get_common_and_uncommon )
import RepoPrefs ( set_lastrepo, write_default_prefs )
import DarcsUtils ( catchall )
#include "impossible.h"
\end{code}
\begin{code}
get_description :: String
get_description =
 "Get a repository.\n"
\end{code}

\options{get}

\haskell{get_help} I recommend using the \verb!--verbose! flag to get, as
this command can take a while, and with no feedback, that can be rather
boring.

If the remote repo and the current directory are in the same filesystem and
that filesystem supports hard links, get will create hard links for the
patch files, which means that the additional storage space needed will be
minimal.  This is \emph{very} good for your disk usage (and for the speed
of running get), so if you want multiple copies of a repository, I strongly
recommend first running \verb!darcs get! to get yourself one copy, and then
running \verb!darcs get! on that copy to make any more you like.  The only
catch is that the first time you run \verb!darcs push! or \verb!darcs pull!
from any of these second copies, by default they will access your first
copy---which may not be what you want.

\begin{code}
get_help :: String
get_help =
 "Get is used to get a local copy of a repository.\n"
\end{code}
\begin{code}
get :: DarcsCommand
get = DarcsCommand {command_name = "get",
                    command_help = get_help,
                    command_description = get_description,
                    command_extra_args = 1,
                    command_extra_arg_help = ["<REPOSITORY>"],
                    command_command = get_cmd,
                    command_prereq = \_ -> return $ Just "",
                    command_get_arg_possibilities = return [],
                    command_argdefaults = nodefaults,
                    command_darcsoptions = [reponame,tagname,
                                            patchname_option,partial,
                                            any_verbosity]}
\end{code}
\begin{code}
get_cmd :: [DarcsFlag] -> [String] -> IO ()
get_cmd opts [inrepodir] = do
  former_dir <- getCurrentDirectory
  repodir <- absolute_dir inrepodir
  repovalid <- is_repo repodir
  unless repovalid $ do putStr $ "Bad repo directory: "++repodir++"\n"
                        exitWith $ ExitFailure 1
  patches_raw <- read_repo repodir
  patches <- pick_patches opts patches_raw
  when (Partial `elem` opts) $ putVerbose "Reading checkpoint...\n"
  mch <- get_checkpoint opts repodir
  myname <- make_repo_name opts repodir
  createDirectory $ myname
  setCurrentDirectory $ former_dir++"/"++myname
  createDirectory "_darcs"
  createDirectory "_darcs/patches"
  createDirectory "_darcs/current"
  createDirectory "_darcs/checkpoints"
  createDirectory "_darcs/prefs"
  writeFile "_darcs/is_converted" "yes" -- FIXME temporary hack
  write_default_prefs
  when (inrepodir == repodir) $ set_lastrepo repodir
  putVerbose $ "Getting the inventory...\n"
  write_inventory "." patches
  copy_repo_patches opts repodir "."
  get_rid_of_unwanted_patches opts
  local_patches <- read_repo "."
  repo_is_local <- doesDirectoryExist repodir
  if repo_is_local && not (chose_specific opts) && not (Partial `elem` opts)
     then do
       s <- slurp_recorded repodir
       copyFileOrUrl (repodir++"/_darcs/prefs/prefs") "_darcs/prefs/prefs"
          `catchall` return ()
       slurp_write s
       s' <- slurp_recorded repodir
       setCurrentDirectory "_darcs/current"
       slurp_write s'
     else do
       setCurrentDirectory $ former_dir++"/"++myname
       putInfo $ "Applying patches to the \"working\" directory...\n"
       if Partial `elem` opts && mch /= Nothing
          then let p_ch = fromJust mch
                   pi_ch = fromJust $ patch2patchinfo p_ch
                   needed_patches = dropWhile ((/= pi_ch).fst) $
                                    reverse $ concat local_patches
                   in do write_checkpoint_patch p_ch
                         case apply_to_slurpy p_ch empty_slurpy of
                             Just s -> slurp_write_dirty s
                             Nothing -> fail "Bad checkpoint!"
                         apply_patches putVorDot putInfo needed_patches
          else apply_patches putVorDot putInfo
                   $ reverse $ concat local_patches
       finishDots
       s <- mmap_slurp_all_but_darcs "."
       setCurrentDirectory "_darcs/current"
       slurp_write s
  putStr $ "Finished getting.\n"
      where putVerbose s = when (Verbose `elem` opts) $ putStr s
            putInfo s = when (not $ Quiet `elem` opts) $ putStr s
            putVorDot s = if Verbose `elem` opts
                          then putStr s
                          else if Quiet `elem` opts
                               then return ()
                               else do putStr "."
                                       hFlush stdout
            finishDots = when (not $ Verbose `elem` opts || Quiet `elem` opts) $
                         putStr "\n"
get_cmd _ _ = impossible
\end{code}

\begin{code}
make_repo_name :: [DarcsFlag] -> FilePath -> IO String
make_repo_name (RepoName n:_) _ = modify_repo_name n (-1)
make_repo_name (_:as) d = make_repo_name as d
make_repo_name [] d =
  case dropWhile (=='.') $ reverse $
       takeWhile (\c -> c /= '/' && c /= ':') $
       dropWhile (=='/') $ reverse d of
  "" -> modify_repo_name "anonymous_repo" (-1)
  base -> modify_repo_name base (-1)

modify_repo_name :: String -> Int -> IO String
modify_repo_name n i = do
    exists <- doesDirectoryExist thename
    file_exists <- doesFileExist thename
    if not exists && not file_exists
       then do when (i /= -1) $
                    putStr $ "Directory '"++ n ++
                           "' already exists, creating repository as '"++
                           thename ++"'\n"
               return thename
       else modify_repo_name n $ i+1
    where thename = if i == -1 then n else n++"_"++show i
\end{code}

If you want to get a specific version of a repository, you have a couple of
options.  The first is to use the \verb!--tag-name! option, which allows
you to specify a specific version of the repository by its tag, and get
precisely that version.  The argument is a regular expression, and get will
grab the most recent tag which matches that regular expression.  The other
option is the \verb!--patch-name! option, which will allow you to get that
patch (i.e.\ the most recent one matching the specified regular expression)
and all older patches.  Note that when specifying a \verb!--patch-name!,
you may get a version of your code that has never before been seen, if the
patches have gotten themselves reordered.  If you ever want to be able to
precisely reproduce a given version, you need to tag it.

\begin{code}
chose_specific :: [DarcsFlag] -> Bool
chose_specific (TagName _:_) = True
chose_specific (PatchName _:_) = True
chose_specific (_:as) = chose_specific as
chose_specific [] = False

pick_patches :: [DarcsFlag] -> PatchSet -> IO PatchSet
pick_patches (TagName t:_) = pictag (mymatch t)
pick_patches (PatchName pn:_) = picpatch (mymatch pn)
pick_patches (_:fs) = pick_patches fs
pick_patches [] = \ ps -> return ps

pictag :: (String -> Bool) -> PatchSet -> IO PatchSet
pictag t (((pinfo,mp):ps):r)
    | take 4 (just_name pinfo) == "TAG " && t (drop 4 (just_name pinfo))
        = do putStrLn $ "Getting tagged version:\n"++show pinfo
             return $ ((pinfo,mp):ps):r
    | otherwise = pictag t (ps:r)
pictag t ([]:r) = pictag t r
pictag _ [] = do putStr "Couldn't find matching tag!\n"
                 return []

picpatch :: (String -> Bool) -> PatchSet -> IO PatchSet
picpatch t (((pinfo,mp):ps):r) =
    if t (just_name pinfo) then do putStr "Getting up to patch:\n"
                                   putStrLn $ show pinfo
                                   return $ ((pinfo,mp):ps):r
                           else picpatch t (ps:r)
picpatch t ([]:r) = picpatch t r
picpatch _ [] = do putStr "Couldn't find matching patch!\n"
                   return []

mymatch :: String -> String -> Bool
mymatch r s = isJust $ matchRegex (mkRegex r) s
\end{code}

\begin{code}
get_rid_of_unwanted_patches :: [DarcsFlag] -> IO ()
get_rid_of_unwanted_patches (TagName _:_) = do
    patches <- read_repo "."
    (_,us',_) <- return $
                 get_common_and_uncommon (patches, [[head $ head patches]])
    when (us' /= []) $ putStr $ "Debug message (not error)-getting rid of:\n"
    when (us' /= []) $ putStr $ unlines $ map (show.fst) $ reverse $ concat us'
    mapM_ (eliminate_patch.fst) $ reverse $ concat us'
get_rid_of_unwanted_patches (_:as) = get_rid_of_unwanted_patches as
get_rid_of_unwanted_patches [] = return ()

eliminate_patch :: PatchInfo -> IO ()
eliminate_patch pinfo = do
  patches <- read_repo "."
  efp $ dropWhile ((/=pinfo).fst) $ reverse $ concat patches
    where efp ((pia,mpa):(pib,mpb):ps) =
              case commute (fromJust mpb, fromJust mpa) of
              Just (pa',pb') ->
                  do writePatch ("_darcs/patches/"++make_filename pib) pb'
                     efp ((pia,Just pa'):ps)
              Nothing -> bug "Noncommuting patch in efp in Get."
          efp [(pin,_)] =
              do patches <- read_repo "."
                 write_inventory "." [just_drop_pi pin $ concat patches]
          efp [] = return ()

just_drop_pi :: PatchInfo -> [(PatchInfo,Maybe Patch)]
              -> [(PatchInfo,Maybe Patch)]
just_drop_pi pidrop ((pinfo,mp):ps)
    | pinfo == pidrop = just_drop_pi pidrop ps
    | otherwise = (pinfo,mp): just_drop_pi pidrop ps
just_drop_pi _ [] = []
\end{code}

