%  Copyright (C) 2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software Foundation,
%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
\begin{code}
module Match ( slurp_match,
               match_patchset, match_second_patchset,
               match_patch,
               get_match, get_second_match,
               one_match, two_matches,
             ) where

import Workaround ( mkRegex, matchRegex )
import Maybe ( isJust )

import PatchInfo ( just_name, human_friendly )
import Patch ( Patch, apply_to_slurpy, invert )
import SlurpDirectory ( Slurpy, slurp_write, slurp_write_and_read_dirty, )
import Repository ( PatchSet, slurp_recorded, read_repo, )
import Depends ( get_patches_in_tag )

import DarcsArguments ( DarcsFlag( MatchPattern, TagName, PatchName ) )
import PatchMatch ( Matcher, match_pattern )
#include "impossible.h"
\end{code}

\begin{code}
slurp_match :: [DarcsFlag] -> IO Slurpy
slurp_match fs =
    case first_matcher fs of
    Nothing -> fail "No pattern specified in slurp_match."
    Just m -> if first_matcher_is_tag fs
              then bug "FIXME: slurp_match isn't correct on tags..."
              else slurp_matcher m

one_match :: [DarcsFlag] -> Bool
one_match fs = isJust $ first_matcher fs

get_match :: FilePath -> [DarcsFlag] -> IO ()
get_match r fs =
    case first_matcher fs of
    Nothing -> fail "Pattern not specified in get_match."
    Just m -> if first_matcher_is_tag fs
              then bug "FIXME: get_match isn't correct on tags..."
              else get_matcher r m

two_matches :: [DarcsFlag] -> Bool
two_matches fs = isJust $ second_matcher fs

get_second_match :: FilePath -> [DarcsFlag] -> IO ()
get_second_match r fs =
    case second_matcher fs of
    Nothing -> fail "Two patterns not specified in get_second_match."
    Just m -> if second_matcher_is_tag fs
              then bug "FIXME: get_second_match isn't correct on tags..."
              else get_matcher r m
\end{code}

\begin{code}
tagmatch :: String -> Matcher
tagmatch r (pinfo,_) =
    take 4 n == "TAG " && isJust (matchRegex (mkRegex r) $ drop 4 n)
    where n = just_name pinfo

mymatch :: String -> Matcher
mymatch r (pinfo,_) = isJust $ matchRegex (mkRegex r) $ just_name pinfo

first_matcher :: [DarcsFlag] -> Maybe Matcher
first_matcher [] = Nothing 
first_matcher (MatchPattern m:_) = Just $ match_pattern m
first_matcher (TagName t:_) = Just $ tagmatch t
first_matcher (PatchName p:_) = Just $ mymatch p
first_matcher (_:fs) = first_matcher fs

first_matcher_is_tag :: [DarcsFlag] -> Bool
first_matcher_is_tag [] = False
first_matcher_is_tag (MatchPattern _:_) = False
first_matcher_is_tag (TagName _:_) = True
first_matcher_is_tag (PatchName _:_) = False
first_matcher_is_tag (_:fs) = first_matcher_is_tag fs

second_matcher :: [DarcsFlag] -> Maybe Matcher
second_matcher [] = Nothing 
second_matcher (MatchPattern _:fs) = first_matcher fs
second_matcher (TagName _:fs) = first_matcher fs
second_matcher (PatchName _:fs) = first_matcher fs
second_matcher (_:fs) = second_matcher fs

second_matcher_is_tag :: [DarcsFlag] -> Bool
second_matcher_is_tag [] = False
second_matcher_is_tag (MatchPattern _:fs) = first_matcher_is_tag fs
second_matcher_is_tag (TagName _:fs) = first_matcher_is_tag fs
second_matcher_is_tag (PatchName _:fs) = first_matcher_is_tag fs
second_matcher_is_tag (_:fs) = second_matcher_is_tag fs
\end{code}

\begin{code}
match_patch :: [DarcsFlag] -> PatchSet -> Patch
match_patch fs ps = case first_matcher fs of
                    Nothing -> bug "Couldn't match_patch"
                    Just m -> find_a_patch m ps

match_patchset :: [DarcsFlag] -> PatchSet -> PatchSet
match_patchset fs ps = case first_matcher fs of
                       Nothing -> bug "Couldn't match_patchset"
                       Just m -> if first_matcher_is_tag fs
                                 then get_matching_tag m ps
                                 else match_a_patchset m ps
match_second_patchset :: [DarcsFlag] -> PatchSet -> PatchSet
match_second_patchset fs ps =
    case second_matcher fs of
    Nothing -> bug "Couldn't match_second_patchset"
    Just m -> if second_matcher_is_tag fs
              then get_matching_tag m ps
              else match_a_patchset m ps

find_a_patch :: Matcher -> PatchSet -> Patch
find_a_patch _ [] = bug "Couldn't find_a_patch"
find_a_patch m ([]:xs) = find_a_patch m xs
find_a_patch m ((p:ps):xs) | m p = the_p
                           | otherwise = find_a_patch m (ps:xs)
    where the_p = case p of
                  (_, Just foo) -> foo
                  (pinf, Nothing) -> error $ "Couldn't read patch " ++
                                     human_friendly pinf

match_a_patchset :: Matcher -> PatchSet -> PatchSet
match_a_patchset _ [] = bug "Couldn't match_a_patchset"
match_a_patchset m ([]:xs) = match_a_patchset m xs
match_a_patchset m ((p:ps):xs) | m p = ((p:ps):xs)
                               | otherwise = match_a_patchset m (ps:xs)

get_matching_tag :: Matcher -> PatchSet -> PatchSet
get_matching_tag _ [] = bug "Couldn't get_matching_tag"
get_matching_tag m ([]:xs) = get_matching_tag m xs
get_matching_tag m xxx@((p:ps):xs)
    | m p = case p of (pinf,_) -> get_patches_in_tag pinf xxx
    | otherwise = get_matching_tag m (ps:xs)
\end{code}

\begin{code}
slurp_matcher :: Matcher -> IO Slurpy
slurp_matcher m = do s <- slurp_recorded "."
                     repo <- read_repo "."
                     if match_exists m repo
                        then return $ apply_foo repo s
                        else fail "Couldn't match pattern."
    where apply_foo [] _ = impossible
          apply_foo ([]:xs) s = apply_foo xs s
          apply_foo ((p:ps):xs) s
              | m p = s
              | otherwise =
                  case snd p of
                  Nothing -> error $ "Sorry, partial repository problem.\n" ++
                             "If you think what you're trying to do is ok"++
                             "then\nreport this as a bug on the " ++
                             "darcs-user list."
                  Just pa ->
                      case apply_to_slurpy (invert pa) s of
                      Just s' -> apply_foo (ps:xs) s'
                      Nothing -> error $ "Inverse patch failed!\n" ++
                                 show (human_friendly $ fst p)

match_exists :: Matcher -> PatchSet -> Bool
match_exists _ [] = False
match_exists m ([]:xs) = match_exists m xs
match_exists m ((p:ps):xs) | m p = True
                           | otherwise = match_exists m (ps:xs)
\end{code}

\begin{code}
get_matcher :: String -> Matcher -> IO ()
get_matcher r m = do s <- slurp_recorded r
                     repo <- read_repo r
                     if match_exists m repo
                        then do slurp_write s
                                apply_foo repo s
                        else fail "Couldn't match pattern."
    where apply_foo [] _ = impossible
          apply_foo ([]:xs) s = apply_foo xs s
          apply_foo ((p:ps):xs) s
              | m p = return ()
              | otherwise =
                  case snd p of
                  Nothing -> error $ "Sorry, partial repository problem.\n" ++
                             "If you think what you're trying to do is ok"++
                             "then\nreport this as a bug on the " ++
                             "darcs-user list."
                  Just pa ->
                      case apply_to_slurpy (invert pa) s of
                      Just s' -> do s'' <- slurp_write_and_read_dirty s'
                                    apply_foo (ps:xs) s''
                      Nothing -> error $ "Inverse patch failed!\n" ++
                                 show (human_friendly $ fst p)
\end{code}
