
A Document is at heart ShowS from the prelude
http://www.haskell.org/onlinereport/standard-prelude.html#$tShowS

Essentially, if you give a Doc a string it'll print out whatever it
wants followed by that string. So (text "foo") makes the Doc that prints
"foo" followed by its argument. The combinator names are taken from
Text.PrettyPrint.HughesPJ, although the behaviour of the two libraries
is slightly different.

The advantage of Printer over simple string appending/concatenating is
that the appends end up associating to the right, e.g.:

  (text "foo" <> text "bar") <> (text "baz" <> text "quux") ""
= \s -> (text "foo" <> text "bar") ((text "baz" <> text "quux") s) ""
= (text "foo" <> text "bar") ((text "baz" <> text "quux") "")
= (\s -> (text "foo") (text "bar" s)) ((text "baz" <> text "quux") "")
= text "foo" (text "bar" ((text "baz" <> text "quux") ""))
= (\s -> "foo" ++ s) (text "bar" ((text "baz" <> text "quux") ""))
= "foo" ++ (text "bar" ((text "baz" <> text "quux") ""))
= "foo" ++ ("bar" ++ ((text "baz" <> text "quux") ""))
= "foo" ++ ("bar" ++ ((\s -> text "baz" (text "quux" s)) ""))
= "foo" ++ ("bar" ++ (text "baz" (text "quux" "")))
= "foo" ++ ("bar" ++ ("baz" ++ (text "quux" "")))
= "foo" ++ ("bar" ++ ("baz" ++ ("quux" ++ "")))

The Empty alternative comes in because you want
    text "a" $$ vcat xs $$ text "b"
($$ means "above", vcat is the list version of $$) to be "a\nb" when xs
is [], but without the concept of an Empty Document each $$ would add a
'\n' and you'd end up with "a\n\nb". Note that Empty /= text "" (the
latter would cause two '\n's).

\begin{code}
module Printer (Doc, Printers(..), Printer, renderWith,
                blueText, text,
                colourPrinters, escapedPrinters, simplePrinters,
                empty, (<>), (<+>), ($$), vcat) where

import Control.Monad.Reader (Reader, runReader, ask)
import Char (isAscii, isPrint, isSpace, ord, intToDigit)
import External ( getTermNColors )

type Doc = Reader Printers Document

data Printers = Printers {defP :: !Printer,
                          blueP :: !Printer}
type Printer = String -> Doc

data Document = Doc (String -> String)
              | Empty

renderWith :: Printers -> Doc -> String
renderWith ps d = case runReader d ps of
                      Empty -> ""
                      Doc f -> f ""

text :: String -> Doc
text x = do ps <- ask
            defP ps x

blueText :: String -> Doc
blueText x = do ps <- ask
                blueP ps x

colourPrinters :: IO Printers

realColourPrinters, escapedPrinters, simplePrinters :: Printers
realColourPrinters  = Printers { defP = escapedPrinter, blueP = bluePrinter }
escapedPrinters = Printers { defP = escapedPrinter, blueP = escapedPrinter }
simplePrinters  = Printers { defP = simplePrinter,  blueP = simplePrinter }

escapedPrinter, simplePrinter, bluePrinter :: Printer
escapedPrinter x = doc (\s -> escape s x)
simplePrinter  x = doc (\s -> x ++ s)
bluePrinter    x = doc (make_blue x) 

colourPrinters = do num_colors <- getTermNColors
                    if num_colors > 4 then return realColourPrinters
                                      else return escapedPrinters

infixl 6 <>
infixl 6 <+>
infixl 5 $$

empty :: Doc
empty = return Empty
doc :: (String -> String) -> Doc
doc f = return $ Doc f

(<>), (<+>), ($$) :: Doc -> Doc -> Doc
a <> b = do ad <- a
            case ad of
                Empty -> b
                Doc af -> do bd <- b
                             return $ Doc (\s -> af $ case bd of
                                                          Empty -> s
                                                          Doc bf -> bf s)

a <+> b = do ad <- a
             case ad of
                 Empty -> b
                 Doc af -> do bd <- b
                              return $ Doc (\s -> af $ case bd of
                                                           Empty -> s
                                                           Doc bf -> ' ':bf s)

a $$ b = do ad <- a
            case ad of
                Empty -> b
                Doc af -> do bd <- b
                             return $ Doc (\s -> af $ case bd of
                                                          Empty -> s
                                                          Doc bf -> '\n':bf s)

vcat :: [Doc] -> Doc
vcat [] = empty
vcat ds = foldr1 ($$) ds

-- escape assumes the input is in ['\0'..'\255']
escape :: String -> String -> String
escape s "" = s
escape s ('\x1B':cs) = make_blue "^[" (escape s cs)
escape s (c:cs)
 | isAscii c && (isPrint c || isSpace c) = c : rest
 | otherwise = let (x, y) = quotRem (ord c) 16
               in make_blue ['\\', intToDigit x, intToDigit y] rest
    where rest = escape s cs

make_blue :: String -> String -> String
make_blue x s = "\x1B[01;34m" ++ x ++ "\x1B[00m" ++ s
\end{code}

